# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gio, GObject

from lollypop.define import App, ArtSize, ArtBehaviour, ViewType, Type
from lollypop.utils import set_cursor_type, popup_widget, emit_signal
from lollypop.helper_signals import SignalsHelper, signals_map
from lollypop.helper_gestures import GesturesHelper


class CoverWidgetBase(SignalsHelper):
    """
        Widget showing current album cover
    """

    @signals_map
    def __init__(self, album, view_type):
        """
            Init cover widget
            @param album as Album
            @param view_type as ViewType
        """
        self.set_property("halign", Gtk.Align.START)
        self.set_property("valign", Gtk.Align.CENTER)
        self._album = album
        self._view_type = view_type
        self.__art_size = 1
        self._artwork = Gtk.Image.new()
        self._artwork.show()
        return [
            (App().album_art, "album-artwork-changed",
             "_on_album_artwork_changed")
        ]

    def set_art_size(self, art_size):
        """
            Set cover artwork size
            @param art_size as int
        """
        if self.__art_size == art_size:
            return
        self.__art_size = art_size
        App().art_helper.set_frame(self._artwork,
                                   "small-cover-frame",
                                   self.__art_size,
                                   self.__art_size)
        App().art_helper.set_album_artwork(
                self._album,
                self.__art_size,
                self.__art_size,
                self._artwork.get_scale_factor(),
                ArtBehaviour.CACHE | ArtBehaviour.CROP_SQUARE,
                self._on_album_artwork)

#######################
# PROTECTED           #
#######################
    def _on_album_artwork_changed(self, art, album_id):
        """
            Update cover for album_id
            @param art as Art
            @param album_id as int
        """
        if self._album is None:
            return
        if album_id == self._album.id:
            App().art_helper.set_album_artwork(
                self._album,
                self.__art_size,
                self.__art_size,
                self._artwork.get_scale_factor(),
                ArtBehaviour.CACHE | ArtBehaviour.CROP_SQUARE,
                self._on_album_artwork)

    def _on_album_artwork(self, surface):
        """
            Set album artwork
            @param surface as str
        """
        if surface is None:
            if self.__art_size == ArtSize.BANNER:
                icon_size = Gtk.IconSize.DIALOG
            else:
                icon_size = Gtk.IconSize.DND
            self._artwork.set_from_icon_name("folder-music-symbolic",
                                             icon_size)
        else:
            self._artwork.set_from_surface(surface)


class EditCoverWidget(Gtk.EventBox, CoverWidgetBase, GesturesHelper):
    """
        Widget showing current album cover (edition allowed)
    """

    def __init__(self, album, view_type):
        """
            Init widget
            @param album as Album
            @param view_type as ViewType
        """
        Gtk.EventBox.__init__(self)
        CoverWidgetBase.__init__(self, album, view_type)
        GesturesHelper.__init__(self, self)
        self.connect("realize", set_cursor_type)
        self.add(self._artwork)

#######################
# PROTECTED           #
#######################
    def _on_primary_press_gesture(self, x, y, event):
        """
            Show artwork popover
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        if self._view_type & ViewType.ALBUM:
            from lollypop.widgets_menu import MenuBuilder
            from lollypop.menu_artwork import AlbumArtworkMenu
            menu = Gio.Menu()
            if App().window.folded:
                from lollypop.menu_header import AlbumMenuHeader
                menu.append_item(AlbumMenuHeader(self._album))
            menu_widget = MenuBuilder(menu, False)
            menu_widget.show()
            menu_ext = AlbumArtworkMenu(self._album, self._view_type, False)
            menu_ext.connect("hidden", self.__close_artwork_menu)
            menu_ext.show()
            menu_widget.add_widget(menu_ext, False)
            self._artwork_popup = popup_widget(menu_widget, self,
                                               None, None, None)

#######################
# PRIVATE             #
#######################
    def __close_artwork_menu(self, action, variant):
        if App().window.folded:
            App().window.container.go_back()
        else:
            self._artwork_popup.destroy()


class BrowsableCoverWidget(Gtk.EventBox, CoverWidgetBase, GesturesHelper):
    """
        Widget showing current album cover
    """

    def __init__(self, album, view_type):
        """
            Init cover widget
            @param album as Album
            @param view_type as ViewType
        """
        Gtk.EventBox.__init__(self)
        CoverWidgetBase.__init__(self, album, view_type)
        self.add(self._artwork)
        GesturesHelper.__init__(self, self)
        self.connect("realize", set_cursor_type)

#######################
# PROTECTED           #
#######################
    def _on_primary_press_gesture(self, x, y, event):
        """
            Browse to album
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        App().window.container.show_view([Type.ALBUM], self._album)


class CoverWidget(Gtk.Bin, CoverWidgetBase):
    """
        Widget showing current album cover
    """

    __gsignals__ = {
        "populated": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, album, view_type):
        """
            Init cover widget
            @param album as Album
            @param view_type as ViewType
        """
        Gtk.Bin.__init__(self)
        CoverWidgetBase.__init__(self, album, view_type)
        self.add(self._artwork)

#######################
# PRIVATE             #
#######################
    def _on_album_artwork(self, surface):
        """
            Set album artwork
            @param surface as str
        """
        emit_signal(self, "populated")
        CoverWidgetBase._on_album_artwork(self, surface)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib, Gtk, Gdk

from gettext import gettext as _

from lollypop.define import App, StorageType, CACHE_PATH
from lollypop.objects_track import Track
from lollypop.objects_album import Album
from lollypop.logger import Logger
from lollypop.dialog_apps import AppsDialog


class ActionsMenu(Gio.Menu):
    """
        ActionsMenu menu for album
    """

    def __init__(self, object):
        """
            Init edit menu
            @param object as Album/Track
        """
        Gio.Menu.__init__(self)
        # Ignore genre_ids/artist_ids
        if isinstance(object, Album):
            self.__object = Album(object.id)
        else:
            self.__object = Track(object.id)
        self.__set_save_action()
        if self.__object.storage_type & StorageType.COLLECTION and\
                not GLib.file_test("/app", GLib.FileTest.EXISTS):
            self.__set_open_action()

#######################
# PRIVATE             #
#######################
    def __set_save_action(self):
        """
            Set save action
        """
        if not self.__object.storage_type & (StorageType.SAVED |
                                             StorageType.COLLECTION):
            save_action = Gio.SimpleAction(name="save_album_action")
            App().add_action(save_action)
            save_action.connect("activate",
                                self.__on_save_action_activate,
                                True)
            menu_item = Gio.MenuItem.new(_("Save in collection"),
                                         "app.save_album_action")
            menu_item.set_attribute_value("close", GLib.Variant("b", True))
            self.append_item(menu_item)
        elif self.__object.storage_type & StorageType.SAVED:
            save_action = Gio.SimpleAction(name="remove_album_action")
            App().add_action(save_action)
            save_action.connect("activate",
                                self.__on_save_action_activate,
                                False)
            menu_item = Gio.MenuItem.new(_("Remove from collection"),
                                         "app.remove_album_action")
            menu_item.set_attribute_value("close", GLib.Variant("b", True))
            self.append_item(menu_item)
        if self.__object.is_web:
            clean_action = Gio.SimpleAction(name="clean_album_action")
            App().add_action(clean_action)
            clean_action.connect("activate",
                                 self.__on_clean_action_activate)
            menu_item = Gio.MenuItem.new(_("Clean cache"),
                                         "app.clean_album_action")
            menu_item.set_attribute_value("close", GLib.Variant("b", True))
            self.append_item(menu_item)
        if isinstance(self.__object, Album) and\
                not self.__object.storage_type & StorageType.COLLECTION:
            buy_action = Gio.SimpleAction(name="buy_album_action")
            App().add_action(buy_action)
            buy_action.connect("activate",
                               self.__on_buy_action_activate)
            menu_item = Gio.MenuItem.new(_("Buy this album"),
                                         "app.buy_album_action")
            menu_item.set_attribute_value("close", GLib.Variant("b", True))
            self.append_item(menu_item)

    def __set_open_action(self):
        """
            Set edit action
        """
        open_tag_action = Gio.SimpleAction(name="open_tag_action")
        App().add_action(open_tag_action)
        open_tag_action.connect("activate", self.__on_open_tag_action_activate)
        menu_item = Gio.MenuItem.new(_("Open with…"),
                                     "app.open_tag_action")
        menu_item.set_attribute_value("close", GLib.Variant("b", True))
        self.append_item(menu_item)

    def __on_buy_action_activate(self, action, variant):
        """
            Launch a browser for Qobuz
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        artists = " ".join(self.__object.artists)
        search = "%s %s" % (artists, self.__object.name)
        uri = "https://www.qobuz.com/search?q=%s" % (
            GLib.uri_escape_string(search, None, True))
        Gtk.show_uri_on_window(App().window,
                               uri,
                               Gdk.CURRENT_TIME)

    def __on_clean_action_activate(self, action, variant):
        """
            clean album cache
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        try:
            if isinstance(self.__object, Album):
                tracks = self.__object.tracks
            else:
                tracks = [self.__object]
            for track in tracks:
                escaped = GLib.uri_escape_string(track.uri, None, True)
                f = Gio.File.new_for_path("%s/web_%s" % (CACHE_PATH, escaped))
                if f.query_exists():
                    f.delete(None)
        except Exception as e:
            Logger.error("ActionsMenu::__on_clean_action_activate():", e)

    def __on_save_action_activate(self, action, variant, save):
        """
            Save album to collection
            @param Gio.SimpleAction
            @param GLib.Variant
            @param save as bool
        """
        if isinstance(self.__object, Album):
            self.__object.save(save)
        else:
            album = self.__object.album
            album.save_track(save, self.__object)
        if not save:
            App().tracks.del_non_persistent()
            App().tracks.clean()
            App().albums.clean()
            App().artists.clean()
            App().genres.clean()

    def __on_open_tag_action_activate(self, action, variant):
        """
            Run tag editor
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        try:
            def launch_app(commandline, f):
                if f.query_exists():
                    args = []
                    for item in commandline.split():
                        if item in ["%U", "%u"]:
                            args.append(f.get_uri())
                        elif item in ["%F", "%f"]:
                            args.append(f.get_path())
                        else:
                            args.append(item)
                    commands = [args,
                                ["flatpak-spawn", "--host"] + args]
                    for cmd in commands:
                        try:
                            (pid, stdin, stdout, stderr) = GLib.spawn_async(
                                cmd, flags=GLib.SpawnFlags.SEARCH_PATH |
                                GLib.SpawnFlags.STDOUT_TO_DEV_NULL,
                                standard_input=False,
                                standard_output=False,
                                standard_error=False
                            )
                            GLib.spawn_close_pid(pid)
                            break
                        except Exception as e:
                            Logger.error("ActionsMenu::launch_app(): %s", e)

            def on_response(dialog, response_id, f):
                if response_id == Gtk.ResponseType.OK:
                    if dialog.commandline is not None:
                        launch_app(dialog.commandline, f)
                dialog.destroy()

            f = Gio.File.new_for_uri(self.__object.uri)
            dialog = AppsDialog(f)
            dialog.connect("response", on_response, f)
            dialog.run()
        except Exception as e:
            Logger.error("ActionsMenu::__on_open_tag_action_activate(): %s", e)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, GObject

import json

from lollypop.define import App, GOOGLE_API_ID
from lollypop.utils import get_network_available, get_page_score, emit_signal
from lollypop.logger import Logger


class BaseWebHelper(GObject.Object):
    """
        Base Web Helper
    """

    __gsignals__ = {
        "uri-loaded": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "uri-content-loaded": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    __BAD_SCORE = 1000000

    def __init__(self):
        """
            Init helper
        """
        GObject.Object.__init__(self)

    def get_uri(self, track, cancellable, methods=[]):
        """
            Get helper URI
            @param track as Track
            @param cancellable as Gio.Cancellable
            @param methods as [function]
        """
        if not methods:
            methods = [self.__get_youtube_id]
            if get_network_available("STARTPAGE"):
                methods.append(self.__get_youtube_id_start)
            if get_network_available("DUCKDUCKGO"):
                methods.append(self.__get_youtube_id_duckduck)
        method = methods.pop(0)
        method(track, cancellable, methods)

#######################
# PRIVATE             #
#######################
    def __get_youtube_id(self, track, cancellable, methods):
        """
            Get youtube id
            @param track as Track
            @param cancellable as Gio.Cancellable
            @param methods as [function]
        """
        unescaped = "%s %s" % (track.artists[0],
                               track.name)
        search = GLib.uri_escape_string(
                            unescaped.replace(" ", "+"),
                            None,
                            True)
        key = App().settings.get_value("cs-api-key").get_string()
        uri = "https://www.googleapis.com/youtube/v3/" +\
              "search?part=snippet&q=%s&" % search +\
              "type=video&key=%s&cx=%s" % (key, GOOGLE_API_ID)
        App().task_helper.load_uri_content(uri, cancellable,
                                           self.__on_get_youtube_id,
                                           track, cancellable, methods)

    def __get_youtube_id_start(self, track, cancellable, methods):
        """
            Get youtube id via startpage
            @param track as Track
            @param cancellable as Gio.Cancellable
            @param methods as [function]
            @return youtube id as str
        """
        unescaped = "%s %s" % (track.artists[0],
                               track.name)
        search = GLib.uri_escape_string(
                        unescaped.replace(" ", "+"),
                        None,
                        True)
        uri = "https://www.startpage.com/do/search?query=%s" % search
        App().task_helper.load_uri_content(uri, cancellable,
                                           self.__on_get_youtube_id_start,
                                           track, cancellable, methods)

    def __get_youtube_id_duckduck(self, track, cancellable, methods):
        """
            Get youtube id via duckduckgo
            @param track as Track
            @param cancellable as Gio.Cancellable
            @param methods as [function]
            @return youtube id as str
        """
        unescaped = "%s %s +youtube" % (track.artists[0],
                                        track.name)
        search = GLib.uri_escape_string(
                        unescaped.replace(" ", "+"),
                        None,
                        True)
        uri = "https://duckduckgo.com/lite/?q=%s" % search
        App().task_helper.load_uri_content(uri, cancellable,
                                           self.__on_get_youtube_id_duckduck,
                                           track, cancellable, methods)

    def __emit_uri_loaded(self, youtube_id, track, cancellable, methods):
        """
            Emit uri loaded for youtube_id
            @param youtube_id as bool
            @param track as Track
            @param cancellable as Gio.Cancellable
            @param methods as [function]
        """
        if youtube_id is not None:
            uri = "https://www.youtube.com/watch?v=%s" % youtube_id
            emit_signal(self, "uri-loaded", uri)
        elif methods:
            self.get_uri(track, cancellable, methods)
        else:
            emit_signal(self, "uri-loaded", "")

    def __on_get_youtube_id(self, uri, status, content, track,
                            cancellable, methods):
        """
            Get youtube id or run another method if not found
            @param uri as str
            @param status as bool
            @param content as bytes
            @param track as Track
            @param cancellable as Gio.Cancellable
            @param methods as [function]
        """
        try:
            youtube_id = None
            if status:
                decode = json.loads(content.decode("utf-8"))
                dic = {}
                best = self.__BAD_SCORE
                for i in decode["items"]:
                    score = get_page_score(i["snippet"]["title"],
                                           track.name,
                                           track.artists[0],
                                           track.album.name)
                    if score == -1 or score == best:
                        continue
                    elif score < best:
                        best = score
                    dic[score] = i["id"]["videoId"]
                # Return url from first dic item
                if best != self.__BAD_SCORE:
                    youtube_id = dic[best]
        except:
            Logger.warning("BaseWebHelper::__on_get_youtube_id(): %s", content)
        self.__emit_uri_loaded(youtube_id, track, cancellable, methods)

    def __on_get_youtube_id_start(self, uri, status, content, track,
                                  cancellable, methods):
        """
            Get youtube id or run another method if not found
            @param uri as str
            @param status as bool
            @param content as bytes
            @param track as Track
            @param cancellable as Gio.Cancellable
            @param methods as [function]
        """
        try:
            from bs4 import BeautifulSoup
            youtube_id = None
            html = content.decode("utf-8")
            soup = BeautifulSoup(html, "html.parser")
            ytems = []
            for link in soup.findAll("a"):
                href = link.get("href")
                title = link.get_text()
                if href is None or title is None or\
                        href.find("youtube.com/watch?v") == -1:
                    continue
                youtube_id = href.split("watch?v=")[1]
                ytems.append((youtube_id, title))
            dic = {}
            best = self.__BAD_SCORE
            for (yid, title) in ytems:
                score = get_page_score(title, track.name,
                                       track.artists[0], track.album.name)
                if score == -1 or score == best:
                    continue
                elif score < best:
                    best = score
                dic[score] = yid
            # Return url from first dic item
            if best != self.__BAD_SCORE:
                youtube_id = dic[best]
        except Exception as e:
            print("$ sudo pip3 install beautifulsoup4")
            Logger.warning("BaseWebHelper::__get_youtube_id_start(): %s", e)
        self.__emit_uri_loaded(youtube_id, track, cancellable, methods)

    def __on_get_youtube_id_duckduck(self, uri, status, content, track,
                                     cancellable, methods):
        """
            Get youtube id or run another method if not found
            @param uri as str
            @param status as bool
            @param content as bytes
            @param track as Track
            @param cancellable as Gio.Cancellable
            @param methods as [function]
        """
        try:
            from bs4 import BeautifulSoup
            youtube_id = None
            html = content.decode("utf-8")
            soup = BeautifulSoup(html, "html.parser")
            ytems = []
            for link in soup.findAll("a"):
                href = GLib.uri_unescape_string(link.get("href"), None)
                title = link.get_text()
                if href is None or title is None or\
                        href.find("youtube.com/watch?v") == -1:
                    continue
                youtube_id = href.split("watch?v=")[1]
                ytems.append((youtube_id, title))
            dic = {}
            best = self.__BAD_SCORE
            for (yid, title) in ytems:
                score = get_page_score(title, track.name,
                                       track.artists[0], track.album.name)
                if score == -1 or score == best:
                    continue
                elif score < best:
                    best = score
                dic[score] = yid
            # Return url from first dic item
            if best != self.__BAD_SCORE:
                youtube_id = dic[best]
        except Exception as e:
            print("$ sudo pip3 install beautifulsoup4")
            Logger.warning("BaseWebHelper::__get_youtube_id_duckduck(): %s", e)
        self.__emit_uri_loaded(youtube_id, track, cancellable, methods)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# Copyright (c) 2015 Jean-Philippe Braun <eon@patapon.info>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio

from urllib.parse import urlparse
from lollypop.define import App, StorageType
from lollypop.objects import Base


class Track(Base):
    """
        Represent a track
    """
    DEFAULTS = {"name": "",
                "album_id": None,
                "artist_ids": [],
                "genre_ids": [],
                "popularity": 0,
                "rate": 0,
                "album_name": "",
                "artists": [],
                "genres": [],
                "duration": 0,
                "number": 0,
                "discnumber": 0,
                "discname": "",
                "year": None,
                "timestamp": 0,
                "mtime": 1,
                "loved": False,
                "storage_type": StorageType.COLLECTION,
                "mb_track_id": None,
                "lp_track_id": None,
                "mb_artist_ids": []}

    def __init__(self, track_id=None, album=None):
        """
            Init track
            @param track_id as int
            @param album as Album
        """
        Base.__init__(self, App().tracks)
        self.id = track_id
        self._uri = None
        self.__uri_loaded = False

        if album is None:
            from lollypop.objects_album import Album
            self.__album = Album(self.album_id)
            self.__album.set_tracks([self], False)
        else:
            self.__album = album

    def __del__(self):
        """
            Remove ref cycles
        """
        self.__album = None

    # Used by pickle
    def __getstate__(self):
        self.db = None
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)
        self.db = App().tracks

    def set_album(self, album):
        """
            Set track album
            @param album as Album
        """
        self.__album = album

    def set_uri(self, uri):
        """
            Set uri
            @param uri as string
        """
        self._uri = uri

    def set_number(self, number):
        """
            Set number
            @param number as int
        """
        self._number = number

    def set_name(self, name):
        """
            Set name
            @param name as str
        """
        self._name = name

    def set_loved(self, loved):
        """
            Mark album as loved
            @param loved as bool
        """
        if self.id >= 0:
            App().tracks.set_loved(self.id, loved)
            self.loved = loved

    def get_featuring_artist_ids(self, album_artist_ids):
        """
            Get featuring artist ids
            @return [int]
        """
        artist_ids = self.db.get_artist_ids(self.id)
        return list(set(artist_ids) - set(album_artist_ids))

    def set_preloaded(self):
        """
            Mark track as preloaded
        """
        self.__uri_loaded = True

    @property
    def uri_loaded(self):
        """
            True if tracks uri is loaded
            @return bool
        """
        # We already have a valid URI (jamendo for example)
        if self.uri.startswith("http"):
            return True
        return self.__uri_loaded

    @property
    def is_web(self):
        """
            True if track is a web track
            @return bool
        """
        return not self.storage_type & (StorageType.COLLECTION |
                                        StorageType.EXTERNAL)

    @property
    def is_http(self):
        """
            True if track is a http track
            @return bool
        """
        parsed = urlparse(self.uri)
        return parsed.scheme in ["http", "https"]

    @property
    def position(self):
        """
            Get track position for album
            @return int
        """
        i = 0
        for track_id in self.__album.track_ids:
            if track_id == self.id:
                break
            i += 1
        return i

    @property
    def first(self):
        """
            Is track first for album
            @return bool
        """
        tracks = self.__album.tracks
        return tracks and self.id == tracks[0].id

    @property
    def last(self):
        """
            Is track last for album
            @return bool
        """
        tracks = self.__album.tracks
        return tracks and self.id == tracks[-1].id

    @property
    def title(self):
        """
            Get track name
            Alias to Track.name
        """
        return self.name

    @property
    def uri(self):
        """
            Get track file uri
            @return str
        """
        if self._uri is None:
            self._uri = App().tracks.get_uri(self.id)
        return self._uri

    @property
    def path(self):
        """
            Get track file path
            @return str
        """
        f = Gio.File.new_for_uri(self.uri)
        path = f.get_path()
        if path is None:
            return ""
        return path

    @property
    def album(self):
        """
            Get track"s album
            @return Album
        """
        return self.__album

    @property
    def album_artists(self):
        """
            Get track album artists, can be != than album.artists as track
            may not have any album
            @return str
        """
        if getattr(self, "_album_artists") is None:
            self._album_artists = self.album.artists
        return self._album_artists



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from locale import strxfrm

from lollypop.utils import noaccents

from lollypop.localized import index_of


class FastScroll(Gtk.ScrolledWindow):
    """
        Widget showing letter and allowing fast scroll on click
        Do not call show on widget, not needed
    """

    def __init__(self, listbox, scrolled):
        """
            Init widget
            @param listbox as Gtk.ListBox
            @param scrolled as Gtk.ScrolledWindow
        """
        Gtk.ScrolledWindow.__init__(self)
        self.__main_scrolled = scrolled
        self.__leave_timeout_id = None
        self.get_style_context().add_class("no-border")
        self.set_margin_end(15)
        self.set_vexpand(True)
        self.set_policy(Gtk.PolicyType.NEVER,
                        Gtk.PolicyType.NEVER)
        self.set_property("halign", Gtk.Align.END)
        self.__chars = []
        self.__listbox = listbox
        self.__scrolled = scrolled
        self.__grid = Gtk.Grid()
        self.__grid.set_orientation(Gtk.Orientation.VERTICAL)
        self.__grid.set_property("valign", Gtk.Align.START)
        self.__grid.show()
        eventbox = Gtk.EventBox()
        eventbox.add(self.__grid)
        eventbox.connect("button-press-event", self.__on_button_press_event)
        eventbox.show()
        self.add(eventbox)
        self.__main_scrolled.get_vadjustment().connect(
            "value_changed", self.__on_value_changed)
        self.connect("scroll-event", self.__on_scroll_event)

    def clear(self):
        """
            Clear values
        """
        for child in self.__grid.get_children():
            child.destroy()
        self.hide()

    def clear_chars(self):
        """
            Clear chars
        """
        self.__chars = []

    def add_char(self, c):
        """
            Add a char to widget, will not be shown
            @param c as char
        """
        if c:
            to_add = noaccents(index_of(c)).upper()
            if to_add not in self.__chars:
                self.__chars.append(to_add)

    def populate(self):
        """
            Populate widget based on current chars
        """
        if not self.__chars:
            return
        label = Gtk.Label.new()
        label.set_margin_start(10)
        label.set_markup('<span font="Monospace"><b>%s</b></span>' % "▲")
        label.show()
        self.__grid.add(label)
        for c in sorted(self.__chars, key=strxfrm):
            label = Gtk.Label.new()
            label.set_margin_start(10)
            label.set_markup('<span font="Monospace"><b>%s</b></span>' % c)
            label.set_opacity(0.4)
            label.show()
            self.__grid.add(label)
        label = Gtk.Label.new()
        label.set_margin_start(10)
        label.set_markup('<span font="Monospace"><b>%s</b></span>' % "▼")
        label.show()
        self.__grid.add(label)
        self.__on_value_changed()

#######################
# PRIVATE             #
#######################
    def __set_margin(self):
        """
            Get top non static entry and set margin based on it position
        """
        margin = 0
        for row in self.__listbox.get_children():
            if row.id >= 0:
                values = row.translate_coordinates(self.__main_scrolled, 0, 0)
                if values is not None:
                    margin = values[1] + 5
                if margin < 5:
                    margin = 5
                self.set_margin_top(margin)
                break

    def __check_value_to_mark(self):
        """
            Look at visible listbox range, and mark char as needed
        """
        start = self.__scrolled.get_vadjustment().get_value()
        end = start + self.__scrolled.get_allocated_height()
        start_value = None
        end_value = None
        for row in self.__listbox.get_children():
            if row.id < 0:
                continue
            values = row.translate_coordinates(self.__listbox, 0, 0)
            if values is not None:
                if row.sortname:
                    name = row.sortname
                else:
                    name = row.name
                if values[1] >= start and start_value is None:
                    start_value = name[0]
                elif values[1] <= end:
                    end_value = name[0]
                else:
                    break
        if start_value is not None and end_value is not None:
            self.__mark_values(start_value, end_value)

    def __mark_values(self, start, end):
        """
            Mark values
            @param start as char
            @param end as char
        """
        start = noaccents(index_of(start)).upper()
        end = noaccents(index_of(end)).upper()
        chars = sorted(self.__chars, key=strxfrm)
        start_idx = chars.index(start)
        end_idx = chars.index(end)
        selected = chars[start_idx:end_idx + 1] + ["▲", "▼"]
        for child in self.__grid.get_children():
            label = child.get_text()
            mark = True if label in selected else False
            if mark:
                child.set_opacity(0.9)
                if label == chars[start_idx]:
                    values = child.translate_coordinates(self.__grid, 0, 0)
                    if values is not None:
                        self.get_vadjustment().set_value(values[1])
            else:
                child.set_opacity(0.4)

    def __on_button_press_event(self, eventbox, event):
        """
            Scroll to activated child char
        """
        char = None
        row = None
        for child in self.__grid.get_children():
            allocation = child.get_allocation()
            if allocation.y <= event.y <= allocation.y + allocation.height:
                char = child.get_text()
                break
        if char is not None:
            if char == "▲":
                row = self.__listbox.get_children()[0]
            elif char == "▼":
                row = self.__listbox.get_children()[-1]
            else:
                for child in self.__listbox.get_children():
                    if child.id < 0:
                        continue
                    if noaccents(index_of(child.sortname))[0].upper() == char:
                        row = child
                        break
        if row is not None:
            values = row.translate_coordinates(self.__listbox,
                                               0, 0)
            if values is not None:
                adj = self.__scrolled.get_vadjustment()
                adj.set_value(values[1])

    def __on_scroll_event(self, scrolled, event):
        """
            Pass event to main scrolled
            @param scrolled as Gtk.ScrolledWindow
            @param event as Gdk.EventScroll
        """
        adj = self.__main_scrolled.get_vadjustment()
        adj.set_value(adj.get_value() + (event.delta_y * 50))
        return True

    def __on_value_changed(self, *ignore):
        """
            Show a popover with current letter
        """
        if self.__chars:
            self.show()
            GLib.idle_add(self.__check_value_to_mark)
            GLib.idle_add(self.__set_margin)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from lollypop.utils_album import tracks_to_albums
from lollypop.utils import get_default_storage_type
from lollypop.define import App, ViewType, MARGIN, Type, Size
from lollypop.objects_album import Album
from lollypop.objects_track import Track
from lollypop.widgets_banner_playlist import PlaylistBannerWidget
from lollypop.view_albums_list import AlbumsListView
from lollypop.helper_signals import SignalsHelper, signals_map
from lollypop.helper_size_allocation import SizeAllocationHelper


class PlaylistsView(AlbumsListView, SignalsHelper, SizeAllocationHelper):
    """
        View showing playlists
    """

    @signals_map
    def __init__(self, playlist_id, view_type):
        """
            Init PlaylistView
            @parma playlist_id as int
            @param view_type as ViewType
        """
        AlbumsListView.__init__(self, [], [], view_type |
                                ViewType.SCROLLED |
                                ViewType.OVERLAY)
        SizeAllocationHelper.__init__(self)
        self.__playlist_id = playlist_id
        self.box.set_width(Size.MEDIUM)
        if view_type & ViewType.DND:
            self.dnd_helper.connect("dnd-finished",
                                    self.__on_dnd_finished)
        self.__banner = PlaylistBannerWidget(playlist_id, self)
        self.__banner.show()
        self.add_widget(self.box, self.__banner)
        return [
                (App().playlists, "playlist-track-added",
                 "_on_playlist_track_added"),
                (App().playlists, "playlist-track-removed",
                 "_on_playlist_track_removed"),
                (App().playlists, "playlists-removed", "_on_playlist_removed"),
                (App().playlists, "playlists-renamed", "_on_playlist_renamed"),
                (App().playlists, "playlists-updated", "_on_playlist_updated")
        ]

    def populate(self):
        """
            Populate view
        """
        if App().playlists.get_smart(self.__playlist_id):
            self.__populate_smart()
        else:
            self.__populate()

    @property
    def args(self):
        """
            Get default args for __class__
            @return {}
        """
        return {"playlist_id": self.__playlist_id,
                "view_type": self.view_type}

    @property
    def filtered(self):
        """
            Get filtered children
            @return [Gtk.Widget]
        """
        filtered = []
        for child in self.children:
            filtered.append(child)
            for subchild in child.children:
                filtered.append(subchild)
        return filtered

    @property
    def scroll_shift(self):
        """
            Get scroll shift for y axes
            @return int
        """
        return self.__banner.height + MARGIN

#######################
# PROTECTED           #
#######################
    def _on_playlist_track_added(self, playlists, playlist_id, uri):
        """
            Append track to album list
            @param playlists as Playlists
            @param playlist_id as int
            @param uri as str
        """
        if playlist_id == self.__playlist_id:
            track = Track(App().tracks.get_id_by_uri(uri))
            album = Album(track.album.id)
            album.set_tracks([track])
            self.add_reveal_albums([album])
            self.add_value(album)

    def _on_playlist_track_removed(self, playlists, playlist_id, uri):
        """
            Remove track from album list
            @param playlists as Playlists
            @param playlist_id as int
            @param uri as str
        """
        if playlist_id == self.__playlist_id:
            track = Track(App().tracks.get_id_by_uri(uri))
            for album_row in self.children:
                if album_row.album.id == track.album.id:
                    for track_row in album_row.children:
                        if track_row.track.id == track.id:
                            track_row.destroy()
                            if len(self.children) == 1:
                                album_row.destroy()
                                break

    def _on_playlist_updated(self, playlists, playlist_id):
        """
            Reload view
            @param playlists as Playlists
            @param playlist_id as int
        """
        self.clear()
        self.populate()

    def _on_playlist_removed(self, playlists, playlist_id):
        """
            Go back
            @param playlists as Playlists
            @param playlist_id as int
        """
        App().window.container.go_back()

    def _on_playlist_renamed(self, playlists, playlist_id):
        """
            Update banner
            @param playlists as Playlists
            @param playlist_id as int
        """
        self.__banner.rename(App().playlists.get_name(playlist_id))

    def _on_row_activated(self, row, track):
        """
            Start playback
            @param row as AlbumRow
            @param track_id as int
        """
        if App().player.is_party:
            App().player.load(track)
        else:
            albums = []
            for album_row in self.children:
                albums.append(album_row.album)
            App().player.play_track_for_albums(track, albums)

    def _on_album_removed(self, row):
        """
            Remove album from playlist
            @param row as AlbumRow
        """
        tracks = []
        for track_id in row.album.track_ids:
            tracks.append(Track(track_id))
        App().playlists.remove_tracks(self.__playlist_id,
                                      tracks,
                                      False)

    def _on_track_removed(self, row, track):
        """
            @param row as AlbumRow
            @param track as Track
        """
        row.album.remove_track(track)
        App().playlists.remove_tracks(self.__playlist_id,
                                      [track],
                                      True)
        if not row.children:
            row.destroy()

#######################
# PRIVATE             #
#######################
    def __populate(self):
        """
            Populate view
        """
        def on_load(albums):
            self.add_reveal_albums(albums)
            AlbumsListView.populate(self, albums)

        def load():
            track_ids = []
            if self.__playlist_id == Type.LOVED:
                for track_id in App().tracks.get_loved_track_ids(
                        [],
                        self.storage_type):
                    if track_id not in track_ids:
                        track_ids.append(track_id)
            else:
                for track_id in App().playlists.get_track_ids(
                        self.__playlist_id):
                    if track_id not in track_ids:
                        track_ids.append(track_id)
            return tracks_to_albums(
                [Track(track_id) for track_id in track_ids])

        App().task_helper.run(load, callback=(on_load,))

    def __populate_smart(self):
        """
            Populate view
        """
        def on_load(albums):
            self.banner.spinner.stop()
            self.add_reveal_albums(albums)
            AlbumsListView.populate(self, albums)

        def load():
            request = App().playlists.get_smart_sql(self.__playlist_id)
            # We need to inject skipped/storage_type
            storage_type = get_default_storage_type()
            split = request.split("ORDER BY")
            split[0] += " AND tracks.loved != %s" % Type.NONE
            split[0] += " AND tracks.storage_type&%s " % storage_type
            track_ids = App().db.execute("ORDER BY".join(split))
            return tracks_to_albums(
                [Track(track_id) for track_id in track_ids])

        self.banner.spinner.start()
        App().task_helper.run(load, callback=(on_load,))

    def __on_dnd_finished(self, dnd_helper):
        """
            Save playlist if needed
            @param dnd_helper as DNDHelper
        """
        if self.__playlist_id >= 0:
            uris = []
            for child in self.children:
                for track in child.album.tracks:
                    uris.append(track.uri)
            App().playlists.clear(self.__playlist_id)
            App().playlists.add_uris(self.__playlist_id, uris)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib


class AdaptiveHelper:
    """
        Adaptive support for View
    """

    def __init__(self):
        """
            Init view
        """
        self.__sidebar_id = None
        self.__selection_ids = {"left": [], "right": []}

    def set_sidebar_id(self, sidebar_id):
        """
            Set sidebar id
            @param sidebar_id as int
        """
        self.__sidebar_id = sidebar_id

    def set_selection_ids(self, selection_ids):
        """
            Set selection ids
            @param selection_ids as {"left": [int], "right": [int])
        """
        self.__selection_ids = selection_ids

    def destroy_later(self):
        """
            Delayed destroy
            Allow animations in stack
        """
        def do_destroy():
            self.destroy()
        self.stop()
        if self.args is not None:
            GLib.timeout_add(1000, do_destroy)

    @property
    def sidebar_id(self):
        """te
            Get sidebar id
            @return int
        """
        return self.__sidebar_id

    @property
    def selection_ids(self):
        """
            Get selection ids (sidebar id + extra ids)
            return [int]
        """
        return self.__selection_ids

    @property
    def args(self):
        """
            Get default args for __class__
            @return {}
        """
        return {}



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# Copyright (C) 2010 Jonathan Matthew (replay gain code)
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from gettext import gettext as _

from lollypop.define import App
from lollypop.widgets_combobox import ComboBox


PRESETS = {
           _("Default"):
            (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
           _("Custom"):
            (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
           "Separator": (),
           _("Acoustic"):
            (3, 2, 2, 2, 3, 2, 2, 3, 2, 4, 2, 2, 1, 1, 4, 5, 7, 8),
           _("Bass"):
            (12, 12, 12, 11, 10, 8, 6, 4, 2, 0,
             -1, -2, -3, -5, -6, -8, -8, -8),
           _("Bass and Treble"):
            (8, 7, 6, 4, 1, -2, -4, -3, 0, 2, 4, 6, 7, 8, 9, 9, 10, 10),
           _("Classical"):
            (3, 2, 1, 0, 2, 1, 2, 1, 2, 3, 1, 1, 1, 2, 4, 3, 2, 1),
           _("Club"):
            (0, 0, 0, 1, 2, 3, 4, 5, 5, 5, 5, 4, 2, 1, 0, 0, 0, 0),
           _("Dance"):
            (11, 11, 8, 8, 8, 5, 5, 0, 0, 0, 0, -5, -5, -5, -8, -8, 0, 0),
           _("Disco"):
            (3, 3, 1, 1, 3, 1, 1, 1, 2, 6, 5, 4, 3, 2, 2, 2, 2, 1),
           _("Drum'n'Bass"):
            (3, 4, 3, 2, 2, 1, 0, 0, 1, 3, 5, 3, 2, 1, 2, 2, 1, 2),
           _("Heavy Metal"):
            (4, 3, 2, 3, 6, 6, 6, 6, 6, 5, 4, 3, 3, 3, 2, 2, 2, 1),
           _("Jazz"):
            (0, 1, 2, 2, 3, 1, 2, 0, 0, 2, 1, 2, 4, 3, 3, 2, 1, 0),
           _("Latin"):
            (0, -2, -1, 0, 1, 1, 2, 2, 3, 4, 1, 2, 2, 2, 3, 2, 1, 1),
           _("Metal"):
            (4, 5, 5, 3, 0, -1, -2, -1, 0, 1, 1, 1, 1, 0, -1, -1, -1, -1),
           _("New Age"):
            (3, 1, 3, 2, 2, 2, 3, 2, 0, 2, 4, 1, 3, 2, 4, 2, 1, 1),
           _("Party"):
            (7, 7, 7, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 7),
           _("Pop"):
            (1, -1, -3, 0, 1, 2, 3, 1, 1, 2, 0, -1, -2, 0, 1, 2, 2, 2),
           _("Reggae"):
            (0, 0, 0, -1, -3, -5, -8, -4, 0, 3, 4, 4, 4, 3, 1, 0, 0, 0),
           _("Rock"):
            (3, -3, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0, 1, 2, 3, 4, 5),
           _("Techno"):
            (6, 7, 7, 6, 4, 2, -1, -3, -2, 0, 2, 3, 4, 4, 4, 3, 2, 1),
           _("Vocal"):
            (2, -1, -1, -1, 2, 2, 4, 3, 4, 4, 3, 2, 0, 0, 0, 0, -1, -1)
}


class EqualizerWidget(Gtk.Bin):
    """
        An equalizer manager widget
    """

    def __init__(self):
        """
            Init widget
        """
        Gtk.Bin.__init__(self)
        self.set_property("valign", Gtk.Align.START)
        self.set_property("halign", Gtk.Align.CENTER)
        self.__timeout_id = None
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Lollypop/EqualizerWidget.ui")
        self.__combobox = ComboBox()
        self.__combobox.show()
        self.__combobox.connect("changed", self.__on_combobox_changed)
        builder.get_object("header_box").pack_end(self.__combobox,
                                                  False, False, 0)
        equalizer = App().settings.get_value("equalizer")
        enabled_equalizer = App().settings.get_value("equalizer-enabled")
        if enabled_equalizer:
            builder.get_object("equalizer_checkbox").set_active(True)
        else:
            self.__combobox.set_sensitive(False)
        for i in range(0, 18):
            scale = builder.get_object("scale%s" % i)
            if i < len(equalizer):
                scale.set_value(equalizer[i])
            setattr(self, "__scale%s" % i, scale)
            scale.connect("value-changed", self.__on_scale_value_changed, i)
            scale.set_sensitive(enabled_equalizer)
        self.add(builder.get_object("widget"))
        preset = ()
        for i in App().settings.get_value("equalizer"):
            preset += (i,)
        for key in PRESETS.keys():
            self.__combobox.append(key)
        self.__set_combobox_value()
        builder.connect_signals(self)

    def do_get_preferred_width(self):
        return (250, 600)

#######################
# PROTECTED           #
#######################
    def _on_format_value(self, scale, value):
        """
            Format scale value
            @param scale as Gtk.Scale
            @param value as float
            @return str
        """
        return "%s dB" % value

    def _on_equalizer_checkbox_toggled(self, button):
        """
            Enable/disable equalizer
            @param button as Gtk.ToggleButton
        """
        active = button.get_active()
        App().settings.set_value("equalizer-enabled",
                                 GLib.Variant("b", active))
        for plugin in App().player.plugins:
            plugin.build_audiofilter()
        App().player.reload_track()
        self.__combobox.set_sensitive(active)
        for i in range(0, 18):
            attr = getattr(self, "__scale%s" % i)
            attr.set_sensitive(active)

#######################
# PRIVATE             #
#######################
    def __set_combobox_value(self):
        """
            Set combobox value based on current equalizer
        """
        combo_set = False
        preset = ()
        for i in App().settings.get_value("equalizer"):
            preset += (i,)
        for key in PRESETS.keys():
            if preset == PRESETS[key]:
                self.__combobox.set_active_id(key)
                combo_set = True
                break
        if not combo_set:
            App().settings.set_value("equalizer-custom",
                                     GLib.Variant("ad", preset))
            self.__combobox.set_label(_("Custom"))

    def __save_equalizer(self):
        """
            Save equalizer to gsettings
        """
        self.__timeout_id = None
        preset = []
        for i in range(0, 18):
            attr = getattr(self, "__scale%s" % i)
            preset.append(attr.get_value())
        App().settings.set_value("equalizer", GLib.Variant("ad", preset))
        self.__set_combobox_value()

    def __on_scale_value_changed(self, scale, band):
        """
            Update equalizer
            @param scale as Gtk.Scale
            @param band as int
        """
        for plugin in App().player.plugins:
            plugin.set_equalizer(band, scale.get_value())
        if self.__timeout_id is not None:
            GLib.source_remove(self.__timeout_id)
        self.__timeout_id = GLib.timeout_add(250, self.__save_equalizer)

    def __on_combobox_changed(self, combobox):
        """
            Update check combobox
            @param combobox as Gtk.ComboBoxText
        """
        key = combobox.get_active_id()
        if key == _("Custom"):
            preset = App().settings.get_value("equalizer-custom")
            App().settings.set_value("equalizer", preset)
            for plugin in App().player.plugins:
                plugin.update_equalizer()
            PRESETS[key] = preset
        keys = PRESETS.keys()
        if key in keys:
            values = PRESETS[key]
            i = 0
            for value in values:
                attr = getattr(self, "__scale%s" % i)
                attr.set_value(value)
                i += 1



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject, Gio

from lollypop.define import CACHE_PATH, App
from lollypop.logger import Logger
from lollypop.utils import emit_signal


class WebHelper(GObject.Object):
    """
        Web helper
    """

    __gsignals__ = {
        "loaded": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(self, track, cancellable):
        """
            Init helper
            @param track as Track
            @param cancellable as Gio.Cancellable
        """
        GObject.Object.__init__(self)
        self.__track = track
        self.__cancellable = cancellable
        if App().settings.get_value("invidious-server").get_string():
            from lollypop.helper_web_invidious import InvidiousWebHelper
            self.__helpers = [InvidiousWebHelper()]
        else:
            from lollypop.helper_web_youtube import YouTubeWebHelper
            self.__helpers = [YouTubeWebHelper()]

    def __del__(self):
        self.__track = None

    def load(self):
        """
            Load track URI
        """
        uri = self.__load_from_cache()
        if uri is None:
            self.__load_uri_with_helper()
        else:
            Logger.info("%s loaded from cache", uri)
            self.__load_uri_content_with_helper(uri, None)

    def save(self, uri):
        """
            Save URI to cache
            @param uri as str
        """
        if not self.__track.lp_track_id:
            return
        try:
            f = Gio.File.new_for_path(
                "%s/%s" % (CACHE_PATH, self.__track.lp_track_id))
            fstream = f.replace(None, False,
                                Gio.FileCreateFlags.REPLACE_DESTINATION,
                                None)
            if fstream is not None:
                fstream.write(uri.encode("utf-8"), None)
                fstream.close()
        except Exception as e:
            Logger.error("WebHelper::save(): %s", e)

    @property
    def uri(self):
        """
            Get track URI
            @return str
        """
        return self.__load_from_cache()

#######################
# PRIVATE             #
#######################
    def __load_from_cache(self):
        """
            Load URI from cache
            @return str/None
        """
        if not self.__track.lp_track_id:
            return None
        try:
            f = Gio.File.new_for_path(
                "%s/%s" % (CACHE_PATH, self.__track.lp_track_id))
            if f.query_exists():
                (stats, content, tag) = f.load_contents()
                return content.decode("utf-8")
        except Exception as e:
            Logger.error("WebHelper::__load_from_cache(): %s", e)
        return None

    def __load_uri_with_helper(self):
        """
            Load track with an helper
        """
        if self.__helpers:
            helper = self.__helpers.pop(0)
            helper.connect("uri-loaded", self.__on_uri_loaded)
            helper.get_uri(self.__track, self.__cancellable)
        else:
            emit_signal(self, "loaded", "")

    def __load_uri_content_with_helper(self, uri, helper):
        """
            Load track uri with and helper
            @param uri as str
            @param helper as BaseWebHelper
        """
        if helper is None and self.__helpers:
            helper = self.__helpers.pop(0)
        if helper is not None:
            helper.connect("uri-content-loaded", self.__on_uri_content_loaded)
            helper.get_uri_content(uri, self.__cancellable)
        else:
            emit_signal(self, "loaded", "")

    def __on_uri_content_loaded(self, helper, uri):
        """
            Emit loaded signal with content
            @param helper as BaseWebHelper
            @param uri as str
        """
        emit_signal(self, "loaded", uri)

    def __on_uri_loaded(self, helper, uri):
        """
            Load URI content
            @param helper as BaseWebHelper
            @param uri as str
        """
        if uri:
            self.__load_uri_content_with_helper(uri, helper)
            self.save(uri)
        else:
            emit_signal(self, "loaded", "")



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk

from lollypop.define import App, MARGIN_SMALL


class TypeAheadWidget(Gtk.Revealer):
    """
        Type ahead widget
    """

    def __init__(self):
        """
            Init widget
        """
        Gtk.Revealer.__init__(self)
        box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        box.show()
        box.set_property("margin", MARGIN_SMALL)
        grid = Gtk.Grid.new()
        grid.show()
        grid.set_hexpand(True)
        grid.set_halign(Gtk.Align.CENTER)
        grid.get_style_context().add_class("linked")
        self.__entry = Gtk.Entry.new()
        self.__entry.show()
        self.__entry.connect("activate", self.__on_type_ahead_activate)
        self.__entry.connect("changed", self.__on_type_ahead_changed)
        self.__entry.set_icon_from_icon_name(Gtk.EntryIconPosition.PRIMARY,
                                             "edit-find-symbolic")
        self.__next_button = Gtk.Button.new_from_icon_name(
            "go-down-symbolic", Gtk.IconSize.BUTTON
        )
        self.__next_button.show()
        self.__prev_button = Gtk.Button.new_from_icon_name(
            "go-up-symbolic", Gtk.IconSize.BUTTON
        )
        self.__prev_button.show()
        self.__close_button = Gtk.Button.new_from_icon_name(
            "window-close-symbolic", Gtk.IconSize.BUTTON
        )
        self.__close_button.show()
        self.__close_button.set_halign(Gtk.Align.END)
        self.__next_button.connect(
            "clicked", lambda x: self.__search_next()
        )
        self.__prev_button.connect(
            "clicked", lambda x: self.__search_prev()
        )
        self.__close_button.connect(
            "clicked", lambda x: self.set_reveal_child(False)
        )
        grid.add(self.__entry)
        grid.add(self.__prev_button)
        grid.add(self.__next_button)
        box.add(grid)
        box.add(self.__close_button)
        self.add(box)
        self.__controller = Gtk.EventControllerKey.new(self.__entry)
        self.__controller.connect("key-pressed", self.__on_key_pressed)
        self.__controller.connect("key-released", self.__on_key_released)

    @property
    def entry(self):
        """
            Get popover entry
            @return Gtk.Entry
        """
        return self.__entry

#######################
# PRIVATE             #
#######################
    def __search_prev(self):
        """
            Search previous item
        """
        view = App().window.container.focused_view
        if view is not None:
            view.search_prev(self.__entry.get_text())

    def __search_next(self):
        """
            Search next item
        """
        view = App().window.container.focused_view
        if view is not None:
            view.search_next(self.__entry.get_text())

    def __on_key_pressed(self, event_controller, keyval, keycode, state):
        """
            Handle keys
            @param event_controller as Gtk.EventController
            @param keyval as int
            @param keycode as int
            @param state as Gdk.ModifierType
        """
        if keyval == Gdk.KEY_Up or keyval == Gdk.KEY_Down:
            return True
        elif keyval == Gdk.KEY_Escape:
            App().window.container.show_filter()

    def __on_key_released(self, event_controller, keyval, keycode, state):
        """
            Handle keys
            @param event_controller as Gtk.EventController
            @param keyval as int
            @param keycode as int
            @param state as Gdk.ModifierType
        """
        if keyval == Gdk.KEY_Up:
            self.__search_prev()
        elif keyval == Gdk.KEY_Down:
            self.__search_next()

    def __on_type_ahead_changed(self, entry):
        """
            Filter current widget
            @param entry as Gtk.entry
        """
        view = App().window.container.focused_view
        if view is not None:
            view.search_for_child(entry.get_text())

    def __on_type_ahead_activate(self, entry):
        """
            Activate row
            @param entry as Gtk.Entry
        """
        view = App().window.container.focused_view
        if view is not None:
            view.activate_child()



# Copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib

# from lollypop.utils import get_network_available
from lollypop.define import NetworkAccessACL, App, Type
from lollypop.ws_token import TokenWebService
from lollypop.logger import Logger


class DirectorWebService:
    """
        Manage web services
    """

    def __init__(self):
        """
            Init object
        """
        self.__token_ws = TokenWebService()
        self.__collection_ws = None
        self.__lastfm_ws = None
        self.__librefm_ws = None
        self.__listenbrainz_ws = None
        self.__spotify_timeout_id = None
        App().settings.connect("changed::network-access",
                               self.__on_network_access_acl_changed)
        App().settings.connect("changed::network-access-acl",
                               self.__on_network_access_acl_changed)

    def start(self):
        """
            Start all web services
        """
        if not App().settings.get_value("network-access"):
            network_acl = 0
        else:
            network_acl = App().settings.get_value(
                "network-access-acl").get_int32()
        self.__handle_collection(network_acl)
        self.__handle_lastfm(network_acl)
        self.__handle_listenbrainz(network_acl)

    def stop(self):
        """
            Stop all web services
            @return bool
        """
        stopped = stopping = 0
        if self.__lastfm_ws is not None:
            stopping += 1
            stopped += self.__lastfm_ws.stop()
        if self.__librefm_ws is not None:
            stopping += 1
            stopped += self.__librefm_ws.stop()
        if self.__listenbrainz_ws is not None:
            stopping += 1
            stopped += self.__listenbrainz_ws.stop()
        if self.__collection_ws is not None:
            stopping += 1
            stopped += self.__collection_ws.stop()
        return stopped == stopping

    @property
    def scrobblers(self):
        """
            Get all scrobbling services
            @return [LastFMWebService/ListenBrainzWebService]
        """
        web_services = []
        for ws in [self.__lastfm_ws, self.__librefm_ws,
                   self.__listenbrainz_ws]:
            if ws is not None:
                web_services.append(ws)
        return web_services

    @property
    def lastfm_ws(self):
        """
            Get Last.fm web service
            @return TokenWebService
        """
        return self.__lastfm_ws

    @property
    def token_ws(self):
        """
            Get token web service
            @return TokenWebService
        """
        return self.__token_ws

    @property
    def collection_ws(self):
        """
            Get Collection web service
            @return CollectionWebService
        """
        return self.__collection_ws

#######################
# PRIVATE             #
#######################
    def __handle_collection(self, acl):
        """
            Start/stop collection based on acl
            @param acl as int
        """
        show_album_lists = App().settings.get_value("shown-album-lists")
        if Type.SUGGESTIONS not in show_album_lists:
            return
        start = acl & NetworkAccessACL["YOUTUBE"] and\
            acl & NetworkAccessACL["DATA"]
        if start and self.__collection_ws is None:
            from lollypop.ws_collection import CollectionWebService
            self.__collection_ws = CollectionWebService()
            Logger.info("Collection web service started")
            self.__collection_ws.start()
            if self.__spotify_timeout_id is None:
                self.__spotify_timeout_id = GLib.timeout_add_seconds(
                    3600, self.__collection_ws.start)
        elif not start and self.__collection_ws is not None:
            if self.__spotify_timeout_id is not None:
                GLib.source_remove(self.__spotify_timeout_id)
                self.__spotify_timeout_id = None
            self.__collection_ws.stop()
            self.__collection_ws = None
            Logger.info("Collection web service stopping")

    def __handle_lastfm(self, acl):
        """
            Start/stop Last.fm based on acl
            @param acl as int
        """
        start = acl & NetworkAccessACL["LASTFM"]
        if start and self.__lastfm_ws is None:
            from lollypop.ws_lastfm import LastFMWebService
            self.__lastfm_ws = LastFMWebService("LASTFM")
            self.__lastfm_ws.start()
            Logger.info("Last.fm web service started")
        elif not start and self.__lastfm_ws is not None:
            self.__lastfm_ws = None
            Logger.info("Last.fm web service stopping")
        start = acl & NetworkAccessACL["LIBREFM"]
        if start and self.__librefm_ws is None:
            from lollypop.ws_lastfm import LastFMWebService
            self.__librefm_ws = LastFMWebService("LIBREFM")
            self.__librefm_ws.start()
            Logger.info("Libre.fm web service started")
        elif not start and self.__librefm_ws is not None:
            self.__librefm_ws = None
            Logger.info("Libre.fm web service stopping")

    def __handle_listenbrainz(self, acl):
        """
            Start/stop ListenBrainz based on acl
            @param acl as int
        """
        start = acl & NetworkAccessACL["MUSICBRAINZ"]
        if start and self.__listenbrainz_ws is None:
            from lollypop.ws_listenbrainz import ListenBrainzWebService
            self.__listenbrainz_ws = ListenBrainzWebService()
            self.__listenbrainz_ws.start()
            App().settings.bind("listenbrainz-user-token",
                                self.__listenbrainz_ws,
                                "user_token", 0)
            Logger.info("ListenBrainz web service started")
        elif not start and self.__listenbrainz_ws is not None:
            App().settings.unbind(self.__listenbrainz_ws,
                                  "listenbrainz-user-token")
            self.__listenbrainz_ws = None
            Logger.info("ListenBrainz web service stopping")

    def __on_network_access_acl_changed(self, *ignore):
        """
            Update available webservices
        """
        self.start()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from lollypop.define import App, Size, ViewType
from lollypop.view_information import InformationView
from lollypop.widgets_popover import Popover


class InformationPopover(Popover):
    """
        Popover with artist information
    """

    def __init__(self, minimal=False):
        """
            Init artist infos
            @param minimal as bool
        """
        Popover.__init__(self)
        self.__minimal = minimal
        self.__width = 10
        self.__view = InformationView(ViewType.SCROLLED |
                                      ViewType.SMALL, minimal)
        self.__view.show()
        self.connect("map", self.__on_map)
        self.get_style_context().add_class("padding")
        self.add(self.__view)

    def populate(self, artist_id=None):
        """
            Show information for artists
            @param artist_id as int
        """
        self.__view.populate(artist_id)

    def do_get_preferred_width(self):
        return (self.__width, self.__width)

#######################
# PROTECTED           #
#######################

#######################
# PRIVATE             #
#######################
    def __on_map(self, widget):
        """
            Connect signal and resize
            @param widget as Gtk.Widget
        """
        size = App().window.get_size()
        if self.__minimal:
            self.__width = min(size[0] * 0.95, 500)
            self.set_size_request(self.__width,
                                  min(size[1] * 0.5, 600))
        else:
            self.__width = Size.NORMAL
            self.set_size_request(self.__width,
                                  min(size[1] * 0.7, 800))



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from lollypop.define import App
from lollypop.assistant import Assistant


class FlatpakAssistant(Assistant):
    """
        Lollypop Flatpak package is now using Freedesktop API to gain access
        to directories, we need to revalidate collection
    """

    def __init__(self):
        """
            Init assistant
        """
        self.__rules = [
            {
              "title": _("Authorization needed"),
              "icon_name": "dialog-password-symbolic",
              "markup": _("""
In order to be able to access your collection,
you need to allow <b>Lollypop</b>."""),
              "uri_label": "",
              "uri": None,
              "right_button_label": None,
              "right_button_style": "",
              "left_button_label": _("Next"),
              "left_button_style": "suggested-action"
            },
            {
              "title": _("Update folders"),
              "icon_name": "folder-music-symbolic",
              "markup": _("""
- <b>Open</b> Lollypop <b>preferences</b>
- Go to the <b>Music</b> tab
- <b>Remove</b> all folders from collection
- Then, <b>add</b> them again"""),
              "uri_label": "",
              "uri": None,
              "right_button_label": _("Back"),
              "right_button_style": "",
              "left_button_label": _("Preferences"),
              "left_button_style": "suggested-action"
            },
        ]
        Assistant.__init__(self, self.__rules)
        self.resize(500, 100)
        self.connect("destroy", self.__on_destroy)

#######################
# PRIVATE             #
#######################
    def __on_destroy(self, window):
        """
            Launch settings
            @param window as Gtk.Window
        """
        App().lookup_action("settings").activate(None)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, GObject

from pickle import load
from time import time

from lollypop.player_albums import AlbumsPlayer
from lollypop.player_auto_random import AutoRandomPlayer
from lollypop.player_auto_similar import AutoSimilarPlayer
from lollypop.player_bin import BinPlayer
from lollypop.player_queue import QueuePlayer
from lollypop.player_linear import LinearPlayer
from lollypop.player_shuffle import ShufflePlayer
from lollypop.player_transitions import TransitionsPlayer
from lollypop.logger import Logger
from lollypop.objects_track import Track
from lollypop.define import App, Type, LOLLYPOP_DATA_PATH
from lollypop.utils import emit_signal


class Player(GObject.GObject, AlbumsPlayer, BinPlayer, AutoRandomPlayer,
             AutoSimilarPlayer, QueuePlayer, LinearPlayer,
             ShufflePlayer, TransitionsPlayer):
    """
        Player object used to manage playback and playlists
    """

    __gsignals__ = {
        "current-changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "duration-changed": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "loading-changed": (GObject.SignalFlags.RUN_FIRST, None,
                            (bool, GObject.TYPE_PYOBJECT)),
        "next-changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "prev-changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "seeked": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "status-changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "volume-changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "queue-changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "playback-added": (GObject.SignalFlags.RUN_FIRST, None,
                           (GObject.TYPE_PYOBJECT,)),
        "playback-updated": (GObject.SignalFlags.RUN_FIRST, None,
                             (GObject.TYPE_PYOBJECT,)),
        "playback-setted": (GObject.SignalFlags.RUN_FIRST, None,
                            (GObject.TYPE_PYOBJECT,)),
        "playback-removed": (GObject.SignalFlags.RUN_FIRST, None,
                             (GObject.TYPE_PYOBJECT,)),
        "rate-changed": (GObject.SignalFlags.RUN_FIRST, None, (int, int))
    }

    def __init__(self):
        """
            Init player
        """
        GObject.GObject.__init__(self)
        AlbumsPlayer.__init__(self)
        AutoRandomPlayer.__init__(self)
        AutoSimilarPlayer.__init__(self)
        BinPlayer.__init__(self)
        QueuePlayer.__init__(self)
        LinearPlayer.__init__(self)
        ShufflePlayer.__init__(self)
        TransitionsPlayer.__init__(self)
        self.__stop_after_track_id = None
        App().settings.connect("changed::repeat", self.update_next_prev)

    def load(self, track):
        """
            Stop current track, load track id and play it
            @param track as Track
        """
        if TransitionsPlayer.load(self, track):
            pass
        else:
            BinPlayer.load(self, track)

    def prev(self):
        """
            Play previous track
        """
        if self.position > App().settings.get_value(
                "previous-threshold").get_int32():
            self.seek(0)
            emit_signal(self, "current-changed")
            if not self.is_playing:
                self.play()
        elif self._prev_track.id is not None:
            self.load(self._prev_track)
        else:
            self.stop()

    def next(self):
        """
            Play next track
        """
        self._on_track_finished(self._current_track)
        if self._next_track.id is not None:
            self.load(self._next_track)
        else:
            self.stop()

    def stop_after(self, track_id):
        """
            Tell player to stop after track_id
            @param track_id as int
        """
        self.__stop_after_track_id = track_id

    def get_current_artists(self):
        """
            Get current artist
            @return artist as string
        """
        artist_ids = self._current_track.album.artist_ids
        if artist_ids[0] == Type.COMPILATIONS:
            artists = ", ".join(self._current_track.artists)
        else:
            artists = ", ".join(self._current_track.album_artists)
        return artists

    def restore_state(self):
        """
            Restore player state
        """
        try:
            if App().settings.get_value("save-state"):
                self._current_track = Track(
                    load(open(LOLLYPOP_DATA_PATH + "/track_id.bin", "rb")))
                self.set_queue(load(open(LOLLYPOP_DATA_PATH +
                                         "/queue.bin", "rb")))
                albums = load(open(LOLLYPOP_DATA_PATH + "/Albums.bin", "rb"))
                (is_playing, was_party) = load(open(LOLLYPOP_DATA_PATH +
                                                    "/player.bin", "rb"))
                if self._current_track.uri:
                    if albums:
                        if was_party:
                            # Tips: prevents player from loading albums
                            self._is_party = True
                            App().lookup_action("party").change_state(
                                GLib.Variant("b", True))
                        self.set_albums(albums)
                        # Load track from player albums
                        index = self.album_ids.index(
                            self._current_track.album.id)
                        for track in self._albums[index].tracks:
                            if track.id == self._current_track.id:
                                self._load_track(track)
                                break
                    if is_playing:
                        self.play()
                    else:
                        self.pause()
                    position = load(open(LOLLYPOP_DATA_PATH + "/position.bin",
                                    "rb"))
                    self.seek(position)
                else:
                    Logger.debug("Player::restore_state(): track missing")
        except Exception as e:
            Logger.error("Player::restore_state(): %s" % e)

    def set_party(self, party):
        """
            Set party mode on if party is True
            Play a new random track if not already playing
            @param party as bool
        """
        ShufflePlayer.set_party(self, party)
        self.update_crossfading()

    def set_prev(self):
        """
            Set previous track
        """
        if self._current_track.id is None:
            return
        try:
            if App().settings.get_value("shuffle") or self.is_party:
                prev_track = ShufflePlayer.prev(self)
            else:
                prev_track = LinearPlayer.prev(self)
            self._prev_track = prev_track
            emit_signal(self, "prev-changed")
        except Exception as e:
            Logger.error("Player::set_prev(): %s" % e)

    def set_next(self):
        """
            Play next track
        """
        if self._current_track.id is None:
            return
        if self._current_track.id == self.__stop_after_track_id:
            self._next_track = Track()
            return
        try:
            next_track = QueuePlayer.next(self)
            if next_track.id is None:
                # Diverge current track to restore playback from queue
                diverge_current_track = None
                if self._queue_current_track is not None:
                    diverge_current_track = self._current_track
                    self._current_track = self._queue_current_track
                if App().settings.get_value("shuffle") or self.is_party:
                    next_track = ShufflePlayer.next(self)
                else:
                    next_track = LinearPlayer.next(self)
                # Restore current track
                if diverge_current_track is not None:
                    self._current_track = diverge_current_track
                    self._queue_current_track = None
            self._next_track = next_track
            emit_signal(self, "next-changed")
        except Exception as e:
            Logger.error("Player::set_next(): %s" % e)

    def update_next_prev(self, *ignore):
        """
            Update next/prev
            @param player as Player
        """
        if self._current_track.id is not None:
            if not self.is_party:
                self.set_next()
            self.set_prev()

    @property
    def next_track(self):
        """
            Current track
        """
        return self._next_track

    @property
    def prev_track(self):
        """
            Current track
        """
        return self._prev_track

    @property
    def stop_after_track_id(self):
        """
            Get stop after track id
            @return int
        """
        return self.__stop_after_track_id

#######################
# PROTECTED           #
#######################
    def _on_track_finished(self, track):
        """
            Scrobble track, update last played time and increment popularity
            @param track as Track
        """
        if track.id is None:
            return
        # Track has been played
        # for at least half its duration, or for 4 minutes
        played = time() - self._start_time
        if played >= track.duration / 2000 or played >= 240:
            self.__scrobble(track, self._start_time)
            if track.id >= 0:
                App().tracks.set_listened_at(track.id, int(time()))
                # Increment popularity
                App().tracks.set_more_popular(track.id)
                # In party mode, linear popularity
                if self.is_party:
                    pop_to_add = 1
                # In normal mode, based on tracks count
                else:
                    count = track.album.tracks_count
                    pop_to_add = int(App().albums.max_count / count)
                App().albums.set_more_popular(track.album_id, pop_to_add)

    def _on_stream_start(self, bus, message):
        """
            On stream start, set next and previous track
        """
        if self.is_in_queue(self._current_track.id):
            self.remove_from_queue(self._current_track.id)
        ShufflePlayer._on_stream_start(self, bus, message)
        BinPlayer._on_stream_start(self, bus, message)
        AutoSimilarPlayer._on_stream_start(self, bus, message)
        self.set_next()
        self.set_prev()

#######################
# PRIVATE             #
#######################
    def __scrobble(self, track, finished_start_time):
        """
            Scrobble on lastfm
            @param track as Track
            @param finished_start_time as int
        """
        # Last.fm policy, force it for ListenBrainz too
        if track.duration < 30000:
            return
        for scrobbler in App().ws_director.scrobblers:
            scrobbler.listen(track, int(finished_start_time))



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# Copyright (c) 2020 David Mandelberg
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from lollypop.define import App, Repeat
from lollypop.objects_album import Album
from lollypop.utils import get_default_storage_type


class AutoRandomPlayer:
    """
        Manage playback for AUTO_RANDOM when going to the end
    """

    def __init__(self):
        """
            Init player
        """
        self.connect("next-changed", self.__on_next_changed)

    def next_album(self):
        """
            Get next album to add
            @return Album
        """
        storage_type = get_default_storage_type()
        for album_id in App().albums.get_randoms(storage_type, None, False, 2):
            if album_id != self.current_track.album.id:
                return Album(album_id)
        return None

#######################
# PRIVATE             #
#######################
    def __on_next_changed(self, player):
        """
            Add a new album if playback finished and wanted by user
        """
        if not self._albums:
            return
        if App().settings.get_enum("repeat") != Repeat.AUTO_RANDOM or\
                player.next_track.id is not None:
            return
        album = self.next_album()
        if album:
            self.add_album(album)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from random import sample

from lollypop.helper_signals import SignalsHelper, signals_map
from lollypop.define import App, Type
from lollypop.objects_track import Track
from lollypop.widgets_albums_rounded import RoundedAlbumsWidget


class PlaylistRoundedWidget(RoundedAlbumsWidget, SignalsHelper):
    """
        Playlist widget showing cover for 4 albums
    """

    @signals_map
    def __init__(self, playlist_id, view_type, font_height):
        """
            Init widget
            @param playlist_id as playlist_id
            @param view_type as ViewType
            @param font_height as int
        """
        name = sortname = App().playlists.get_name(playlist_id)
        RoundedAlbumsWidget.__init__(self, playlist_id, name,
                                     sortname, view_type, font_height)
        self._track_ids = []
        self._genre = Type.PLAYLISTS
        return [
            (App().art, "artwork-cleared", "_on_artwork_cleared")
        ]

    def populate(self):
        """
            Populate widget content
        """
        if self._artwork is None:
            RoundedAlbumsWidget.populate(self)
        else:
            self.set_artwork()

    @property
    def track_ids(self):
        """
            Get current track ids
            @return [int]
        """
        return self._track_ids

    @property
    def artwork_name(self):
        """
            Get artwork name
            return str
        """
        return "playlist_" + self.name

#######################
# PROTECTED           #
#######################
    def _get_album_ids(self):
        """
            Get album ids
            @return [int]
        """
        album_ids = []
        if self._data > 0 and App().playlists.get_smart(self._data):
            request = App().playlists.get_smart_sql(self._data)
            if request is not None:
                self._track_ids = App().db.execute(request)
        else:
            self._track_ids = App().playlists.get_track_ids(self._data)
        sample(self._track_ids, len(self._track_ids))
        for track_id in self._track_ids:
            track = Track(track_id)
            if track.album.id not in album_ids:
                album_ids.append(track.album.id)
            if len(album_ids) == self._ALBUMS_COUNT:
                break
        return album_ids

    def _on_artwork_cleared(self, art, name, prefix):
        """
            Update artwork
            @param art as Art
            @param name as str
            @param prefix as str
        """
        if self._artwork is not None:
            self.set_artwork()

#######################
# PRIVATE             #
#######################
